from mitoinstaller.log_utils import log_error
import traceback
from typing import List

from mitoinstaller.installer_steps.installer_step import InstallerStep


def run_installer_steps(installer_steps: List[InstallerStep]):
    """
    A helper function that runs the installer steps, returning
    True if they all terminate without error, and False if any
    of them return with an error.
    """

    for installer_step in installer_steps:
        try:
            # Print the step name, so the user knows what's going on
            print(installer_step.installer_step_name)
            installer_step.execute()
        except:
            # Log the error. Print if it's not an optional step
            log_error(installer_step.installer_step_name, print_to_user=(not installer_step.optional))

            # If we error on a non optional step, then return False
            # and continue if it was an optional step
            if installer_step.optional:
                continue
            else:
                return False
    
    return True


    


import json
import pytest
from tests.conftest import VirtualEnvironment
from mitoinstaller.user_install import USER_JSON_PATH, get_static_user_id

def test_upgrade_mitosheet3_installed_correctly_upgrades_stays_3(venv: VirtualEnvironment):
    venv.run_python_module_command(['pip', 'install', '-r', 'requirements.txt'])
    venv.run_python_module_command(['mitoinstaller', 'install'])
    static_user_id_one = get_static_user_id()

    venv.run_python_module_command(['mitoinstaller', 'upgrade'])
    static_user_id_two = get_static_user_id()

    assert static_user_id_one == static_user_id_two
    
    mitosheet3_version = venv.get_package_version('mitosheet3')
    mitosheet_version = venv.get_package_version('mitosheet')
    jlab_version = venv.get_package_version('jupyterlab')

    assert mitosheet3_version is not None 
    assert mitosheet_version is None 
    assert jlab_version.startswith('3')

def test_upgrade_mitosheet_installed_correctly_upgrades_stays_2(venv: VirtualEnvironment):
    venv.run_python_module_command(['pip', 'install', '-r', 'requirements.txt'])
    venv.run_python_module_command(['pip', 'install', 'jupyterlab<3.0'])
    venv.run_python_module_command(['pip', 'install', 'mitosheet'])
    venv.run_python_module_command(['jupyter', 'labextension', 'install', '@jupyter-widgets/jupyterlab-manager@2'])
    static_user_id_one = get_static_user_id()

    venv.run_python_module_command(['mitoinstaller', 'upgrade'])
    static_user_id_two = get_static_user_id()

    assert static_user_id_one == static_user_id_two

    jlab_version = venv.get_package_version('jupyterlab')
    mitosheet3_version = venv.get_package_version('mitosheet3')
    mitosheet_version = venv.get_package_version('mitosheet')
    assert jlab_version.startswith('2')
    assert mitosheet3_version is None
    assert mitosheet_version is not None

def test_user_json_only_has_static_user_id():
    with open(USER_JSON_PATH, 'w+') as f:
        f.write(json.dumps({
            'static_user_id': 'github_action'
        }))
    from mitoinstaller.user_install import user_json_only_has_static_user_id
    
    assert user_json_only_has_static_user_id()

    with open(USER_JSON_PATH, 'w+') as f:
        f.write(json.dumps({
            'static_user_id': 'github_action',
            'other': 'nah'
        }))

    assert not user_json_only_has_static_user_id()

    



import pytest
from tests.conftest import WINDOWS

from mitoinstaller.commands import run_command


def test_command_suceedes():
    if WINDOWS:
        run_command(['dir'])
    else:
        run_command(['ls'])

def test_command_fails():
    with pytest.raises(Exception) as e:
        run_command(['cd', './fake-folder'])
    assert 'file' in str(e)





import os
import sys

from mitoinstaller.commands import (check_running_jlab_3_processes,
                                    check_running_jlab_processes)
from mitoinstaller.installer_steps.installer_step import InstallerStep
from mitoinstaller.log_utils import log_error
from mitoinstaller.starter_notebook import (MITO_STARTER_NOTEBOOK_PATH,
                                            try_create_starter_notebook)
from mitoinstaller.user_install import is_running_test
from mitoinstaller.create_startup_file import create_startup_file
from termcolor import colored


def replace_process_with_jupyter_lab():
    """
    Switch the currently running process with a running
    Jupyter Lab instance.

    If there is a current JupyterLab process running, or
    if we are running tests, then we do not try and launch
    JLab. Instead, we raise an exception, so that this is
    then logged (only in case of other JLab).
    """
    if is_running_test():
        return
    
    if check_running_jlab_processes():
        raise Exception('Cannot replace process with JupyterLab, currently running existing JLab instances')

    os.execl(sys.executable, 'python', '-m', 'jupyter', 'lab', MITO_STARTER_NOTEBOOK_PATH)


def get_success_message():
    """
    We show a different message depending on if this is an install or an upgrade,
    and we further tell the user different things if they have a currently running
    JLab instance (as they need to restart this)
    """
    running_jlab = check_running_jlab_3_processes()

    separator_line = '----------------------------------------------------------------------------'

    install_or_upgrade = 'install'
    install_start = 'Mito has finished installing'
    # NOTE: this is a bug that gives the wrong message, but it's fine, 
    # since most people are installing and not upgrading.
    upgrade_start = 'Mito has finished installing'

    launch_jlab = 'Launch JupyterLab with:\t' + colored('python -m jupyter lab', 'green')
    relaunch_jlab = colored('Please shut down the currently running JupyterLab and relaunch it to enable Mito', 'green')

    render_mitosheet = colored('Then render a mitosheet following the instructions here: https://docs.trymito.io/how-to/creating-a-mitosheet', 'green')

    if not running_jlab:
        if install_or_upgrade == 'install':
            return '\n{separator_line}\n{install_start}\n\n{launch_jlab}\n\n{render_mitosheet}\n{separator_line}'.format(
                separator_line=separator_line,
                install_start=install_start,
                launch_jlab=launch_jlab,
                render_mitosheet=render_mitosheet,
            )
        else:
            return '\n{separator_line}\n{upgrade_start}\n\n{launch_jlab}\n\n{render_mitosheet}\n{separator_line}'.format(
                separator_line=separator_line,
                upgrade_start=upgrade_start,
                launch_jlab=launch_jlab,
                render_mitosheet=render_mitosheet,
            )
    else:
        if install_or_upgrade == 'install':
            return '\n{separator_line}\n{install_start}\n\n{relaunch_jlab}\n\n{render_mitosheet}\n{separator_line}'.format(
                separator_line=separator_line,
                install_start=install_start,
                relaunch_jlab=relaunch_jlab,
                render_mitosheet=render_mitosheet,
            )
        else:
            return '\n{separator_line}\n{upgrade_start}\n\n{relaunch_jlab}\n\n{render_mitosheet}\n{separator_line}'.format(
                separator_line=separator_line,
                upgrade_start=upgrade_start,
                relaunch_jlab=relaunch_jlab,
                render_mitosheet=render_mitosheet,
            )

def print_success_message():
    """
    Prints a success message to the user, in the case that we cannot
    launch JupyterLab. Furthermore, logs that the jupyter lab instance
    could not be launched.
    """
    # If we get past this, then JLab must have not been able to replace the current process, so we 
    # log that, and then print the success message anyways
    log_error('failed_launch_jupyterlab', print_to_user=False)

    print(get_success_message())


FINAL_INSTALLER_STEPS = [
    InstallerStep(
        'Create import mito startup file',
        create_startup_file,
        True
    ),
    InstallerStep(
        'Creating a Mitosheet starter notebook',
        try_create_starter_notebook
    ),
    InstallerStep(
        'Start JupyterLab',
        replace_process_with_jupyter_lab,
        optional=True
    ),
    # We do out best to replace the running process with JupyterLab, but 
    # if this fails, then just print a success message to the user
    InstallerStep(
        'Finish Installation',
        print_success_message
    )
]



import pytest
from tests.conftest import VirtualEnvironment
import os


def test_install_mitosheet3_no_dependencies(venv: VirtualEnvironment):
    venv.run_python_module_command(['pip', 'install', '-r', 'requirements.txt'])
    venv.run_python_module_command(['mitoinstaller', 'install'])
    mitosheet3_version = venv.get_package_version('mitosheet3')
    mitosheet_version = venv.get_package_version('mitosheet')
    jlab_version = venv.get_package_version('jupyterlab')

    assert mitosheet3_version is not None 
    assert mitosheet_version is None 
    assert jlab_version.startswith('3')


@pytest.mark.parametrize("jupterlab_with_version", ['jupyterlab<1.0', 'jupyterlab<2.0', 'jupyterlab<3.0'])
def test_upgrade_jlab_no_dependencies(venv: VirtualEnvironment, jupterlab_with_version):
    venv.run_python_module_command(['pip', 'install', '-r', 'requirements.txt'])
    venv.run_python_module_command(['pip', 'install', '\"{jupterlab_with_version}\"'.format(jupterlab_with_version=jupterlab_with_version)])
    venv.run_python_module_command(['mitoinstaller', 'install'])
    mitosheet3_version = venv.get_package_version('mitosheet3')
    mitosheet_version = venv.get_package_version('mitosheet')
    jlab_version = venv.get_package_version('jupyterlab')

    assert mitosheet3_version is not None 
    assert mitosheet_version is None 
    assert jlab_version.startswith('3')


@pytest.mark.parametrize("jupterlab_with_version", ['jupyterlab<1.0', 'jupyterlab<2.0', 'jupyterlab<3.0'])
def test_upgrade_jlab_with_mitosheet_dependency(venv: VirtualEnvironment, jupterlab_with_version):
    venv.run_python_module_command(['pip', 'install', '-r', 'requirements.txt'])
    venv.run_python_module_command(['pip', 'install', jupterlab_with_version])
    venv.run_python_module_command(['pip', 'install', 'mitosheet'])
    jlab_version = venv.get_package_version('jupyterlab')
    mitosheet3_version = venv.get_package_version('mitosheet3')
    mitosheet_version = venv.get_package_version('mitosheet')
    assert jlab_version.startswith('2')
    assert mitosheet3_version is None
    assert mitosheet_version is not None

    stdout, stderr = venv.run_python_module_command(['mitoinstaller', 'install'])
    print(stdout, stderr)
    mitosheet3_version = venv.get_package_version('mitosheet3')
    mitosheet_version = venv.get_package_version('mitosheet')
    jlab_version = venv.get_package_version('jupyterlab')

    assert mitosheet3_version is not None 
    assert mitosheet_version is None 
    assert jlab_version.startswith('3')


# Version map taken from https://www.npmjs.com/package/@jupyter-widgets/jupyterlab-manager
@pytest.mark.parametrize("jupterlab_version, dependency", [
    ('0.35.0', '@jupyter-widgets/jupyterlab-manager@0.38'), 
    ('1.2.0', '@jupyter-widgets/jupyterlab-manager@1.1'), 
])
def test_not_upgrade_jlab_with_other_dependency(venv: VirtualEnvironment, jupterlab_version, dependency):
    venv.run_python_module_command(['pip', 'install', '-r', 'requirements.txt'])
    venv.run_python_module_command(['pip', 'install', 'jupyterlab=={jupterlab_version}'.format(jupterlab_version=jupterlab_version)])
    assert venv.get_package_version('jupyterlab') == jupterlab_version

    venv.run_python_module_command(['jupyter', 'labextension', 'install', '{dependency}'.format(dependency=dependency)])
    venv.run_python_module_command(['mitoinstaller', 'install'])
    
    mitosheet3_version = venv.get_package_version('mitosheet3')
    mitosheet_version = venv.get_package_version('mitosheet')
    jlab_version = venv.get_package_version('jupyterlab')

    assert mitosheet3_version is None 
    assert mitosheet_version is None 
    assert jlab_version == jupterlab_version


def test_installs_mitosheet_jlab_2(venv: VirtualEnvironment):
    venv.run_python_module_command(['pip', 'install', '-r', 'requirements.txt'])
    venv.run_python_module_command(['pip', 'install', 'jupyterlab==2.2.5'])
    venv.run_python_module_command(['jupyter', 'labextension', 'install', '@jupyter-widgets/jupyterlab-manager@2'])
    
    stdout, stderr = venv.run_python_module_command(['mitoinstaller', 'install'])
    print(stdout, stderr)
    mitosheet3_version = venv.get_package_version('mitosheet3')
    mitosheet_version = venv.get_package_version('mitosheet')

    assert mitosheet3_version is None 
    assert mitosheet_version is not None 
    assert 'mitosheet' in venv.get_jupyterlab_extension_names()


def test_installs_mitosheet3_jlab_3(venv: VirtualEnvironment):
    venv.run_python_module_command(['pip', 'install', '-r', 'requirements.txt'])
    venv.run_python_module_command(['pip', 'install', 'jupyterlab==3.0'])
    
    venv.run_python_module_command(['mitoinstaller', 'install'])
    
    mitosheet3_version = venv.get_package_version('mitosheet3')
    mitosheet_version = venv.get_package_version('mitosheet')

    assert mitosheet3_version is not None 
    assert mitosheet_version is None 
    assert 'mitosheet3' in venv.get_jupyterlab_extension_names()


def test_create_startup_file(venv: VirtualEnvironment):
    venv.run_python_module_command(['pip', 'install', '-r', 'requirements.txt'])
    venv.run_python_module_command(['pip', 'install', 'jupyterlab==3.0'])
    
    venv.run_python_module_command(['mitoinstaller', 'install'])

    IMPORT_MITOSHEET_FILE_PATH = os.path.join(os.path.expanduser("~"), '.ipython', 'profile_default', 'startup', 'import_mitosheet.py')
    
    assert os.path.exists(IMPORT_MITOSHEET_FILE_PATH) 



"""
The Mito Installer package contains utils for installing
Mito within your Python enviornment.

Long term, we aim to meet:
1. This package has minimal dependencies, both for speed of download and the ultimate portability.
2. The installation attempts to fail as early as possible, and to give the user as much help
   help as possible while doing so.
"""
from colorama import init
from termcolor import colored

from mitoinstaller.do_upgrade_to_jupyterlab_3 import do_upgrade_to_jupyterlab_3
from mitoinstaller.install import do_install_or_upgrade


def main():
    """
    The main function of the Mito installer, this function is responsible
    for either installing mitosheet or upgrading mitosheet.

    To install Mito (for the first time):
    python -m mitoinstaller install

    To upgrade Mito:
    python -m mitoinstaller upgrade
    """
    import sys
    init()

    if len(sys.argv) > 1:
        command = sys.argv[1]
    else:
        command = ''

    # Check if it's a pro install
    is_pro = '--pro' in sys.argv

    if command == 'install':
        do_install_or_upgrade('install', is_pro)
    elif command == 'uninstall':
        print('To uninstall, run,', colored('`pip uninstall mitosheet`', 'green'), 'and', colored('`pip uninstall mitosheet3`', 'green'))
    elif command == 'upgrade':
        do_install_or_upgrade('upgrade', is_pro)
    elif command == 'upgrade_to_jupyterlab_3':
        do_upgrade_to_jupyterlab_3()
    else:
        # NOTE: we don't add upgrade_to_jupyterlab_3 to the help.
        # We only send this command to the users who need to know this (namely, those that need to upgrade)
        print('\nProper usage is', colored('`python -m mitoinstaller install`', 'green'), 'or', colored('`python -m mitoinstaller upgrade`', 'green'), '\n\nTry running the command ', colored('`python -m mitoinstaller install`', 'green'), '\n')
 

if __name__ == '__main__':
    main()



from mitoinstaller.commands import (get_jupyterlab_metadata,
                                    install_pip_packages,
                                    uninstall_pip_packages)
from mitoinstaller.installer_steps.installer_step import InstallerStep
from mitoinstaller.log_utils import log


def install_step_mitosheet3_check_dependencies():
    """
    This is the most complex step in the installation process. It's
    goal is to check if the users existing installation can safely
    be upgraded to JLab 3.0. 

    To do this, it checks a variety of conditions, mostly around what
    version of JLab they have installed, and if this version of JLab has
    any installed dependencies (that we cannot safely upgrade).
    """

    jupyterlab_version, extension_names = get_jupyterlab_metadata()

    # If no JupyterLab is installed, we can continue with install, as
    # there are no conflict dependencies
    if jupyterlab_version is None:
        return
    
    # If JupyterLab 3 is installed, then we are are also good to go
    if jupyterlab_version.startswith('3'):
        return
    
    if len(extension_names) == 0:
        return
    elif len(extension_names) == 1 and extension_names[0] == 'mitosheet':
        uninstall_pip_packages('mitosheet')
        log('uninstalled_mitosheet_labextension',
            {
                'jupyterlab_version': jupyterlab_version,
                'extension_names': extension_names
            }
        )
        return
    
    else:
        raise Exception('Installed extensions {extension_names}'.format(extension_names=extension_names))


def install_step_mitosheet3_install_mitosheet3():
    install_pip_packages('mitosheet3')


MITOSHEET3_INSTALLER_STEPS = [
    InstallerStep(
        'Checking dependencies',
        install_step_mitosheet3_check_dependencies
    ),
    InstallerStep(
        'Installing mitosheet3',
        install_step_mitosheet3_install_mitosheet3
    ),
]



"""
We specify the installer as a list of steps that 
must run in order, for the installer.

Currently, we take the following steps:
1. First, try and install mitosheet3. Because mitosheet3 uses 
   JLab 3, we first check if the user has any conflicting
   JLab 2 dependencies installed. If they do, we abort. If they 
   don't, we update them to JLab 3.
2. Then, if the above installation of mitosheet3 fails for 
   any reason, we try to install mitosheet on JLab 2. 
3. If neither of them work, we give up, sadly. 
"""

from mitoinstaller.commands import exit_after_error
from mitoinstaller.installer_steps import (FINAL_INSTALLER_STEPS,
                                           INITIAL_INSTALLER_STEPS,
                                           MITOSHEET3_INSTALLER_STEPS,
                                           MITOSHEET_INSTALLER_STEPS,
                                           run_installer_steps)
from mitoinstaller.installer_steps.initial_installer_steps import INITIAL_INSTALLER_STEPS_PRO
from mitoinstaller.log_utils import log, log_error


def do_install_or_upgrade(install_or_upgrade, is_pro: bool):
    """
    install_or_upgrade is the workhorse of actually installing mito. 
    It first attempts mitosheet3, and if that fails, then mitosheet.

    install_or_upgrade should be 'install' or 'upgrade'.

    Notably, the process for installing Mito initially and upgrading Mito are
    identical. As such, we reuse this function to upgrade, just with different
    error and logging messages.
    """
    print("Starting {install_or_upgrade}...".format(install_or_upgrade=install_or_upgrade))

    # Change the installation depending if the user is a pro user or not
    if not is_pro:
        run_installer_steps(INITIAL_INSTALLER_STEPS)
    else:
        run_installer_steps(INITIAL_INSTALLER_STEPS_PRO)

    # First, we try to install mitosheet3
    installed_mitosheet_3 = run_installer_steps(MITOSHEET3_INSTALLER_STEPS)
    if installed_mitosheet_3:
        # If we installed mitoshet3, log that we did
        log('install_finished_mitosheet3')
        
        # Then run the final installer steps and exit
        run_installer_steps(FINAL_INSTALLER_STEPS)
        exit(0)
    else:
        # Otherwise, log the error (which prints to the user). Note that we don't give up here!
        log_error('install_failed_mitosheet3', print_to_user=False)

    # Then, try to install mitosheet
    installed_mitosheet = run_installer_steps(MITOSHEET_INSTALLER_STEPS)
    if installed_mitosheet:
        log('install_finished_mitosheet')
    else:
        # Log the error and give up if both our installations failed
        log_error('install_failed_mitosheet', print_to_user=False)
        # TODO: we could make a clickable link to debugging instructions, or an email to send to Jake?
        # I think we should prompt users to do this, defaulting to Yes!
        exit_after_error(install_or_upgrade)

    # Then, run the final installer steps
    run_installer_steps(FINAL_INSTALLER_STEPS)

    



import pytest
from tests.conftest import VirtualEnvironment

def test_allows_upgrade_to_jlab3_when_mitosheet_installed(venv: VirtualEnvironment):
    venv.run_python_module_command(['pip', 'install', '-r', 'requirements.txt'])
    venv.run_python_module_command(['pip', 'install', 'mitosheet'])
    venv.run_python_script('import mitosheet')
    venv.run_python_module_command(['jupyter', 'labextension', 'install', '@jupyter-widgets/jupyterlab-manager@2'])

    stdout, stderr = venv.run_python_module_command(['mitoinstaller', 'upgrade_to_jupyterlab_3'])
    print(stdout, stderr)

    mitosheet3_version = venv.get_package_version('mitosheet3')
    mitosheet_version = venv.get_package_version('mitosheet')
    jlab_version = venv.get_package_version('jupyterlab')

    assert mitosheet3_version is not None 
    assert mitosheet_version is None 
    assert jlab_version.startswith('3')


def test_upgrade_to_jlab3_fails_when_other_extension(venv: VirtualEnvironment):
    venv.run_python_module_command(['pip', 'install', '-r', 'requirements.txt'])
    venv.run_python_module_command(['pip', 'install', 'mitosheet'])
    venv.run_python_module_command(['jupyter', 'labextension', 'install', '@jupyter-widgets/jupyterlab-manager@2'])
    venv.run_python_module_command(['jupyter', 'labextension', 'install', 'jupyterlab-drawio'])

    venv.run_python_module_command(['mitoinstaller', 'upgrade_to_jupyterlab_3'])

    mitosheet3_version = venv.get_package_version('mitosheet3')
    mitosheet_version = venv.get_package_version('mitosheet')
    jlab_version = venv.get_package_version('jupyterlab')

    assert mitosheet3_version is None 
    assert mitosheet_version is not None 
    assert jlab_version.startswith('2')

def test_upgrade_to_jlab3_fails_when_no_mitosheet_extension(venv: VirtualEnvironment):
    venv.run_python_module_command(['pip', 'install', '-r', 'requirements.txt'])
    venv.run_python_module_command(['pip', 'install', 'jupyterlab<3.0'])

    venv.run_python_module_command(['mitoinstaller', 'upgrade_to_jupyterlab_3'])

    mitosheet3_version = venv.get_package_version('mitosheet3')
    mitosheet_version = venv.get_package_version('mitosheet')
    jlab_version = venv.get_package_version('jupyterlab')

    assert mitosheet3_version is None 
    assert mitosheet_version is None 
    assert jlab_version.startswith('2')


def test_upgrade_to_jlab3_fails_when_already_on_jlab3(venv: VirtualEnvironment):
    venv.run_python_module_command(['pip', 'install', '-r', 'requirements.txt'])
    venv.run_python_module_command(['mitoinstaller', 'install'])

    mitosheet3_version = venv.get_package_version('mitosheet3')
    mitosheet_version = venv.get_package_version('mitosheet')
    jlab_version = venv.get_package_version('jupyterlab')

    assert mitosheet3_version is not None 
    assert mitosheet_version is None 
    assert jlab_version.startswith('3')


from mitoinstaller import __version__
from mitoinstaller.commands import upgrade_mito_installer
from mitoinstaller.installer_steps.installer_step import InstallerStep
from mitoinstaller.log_utils import identify, log
from mitoinstaller.user_install import (get_static_user_id,
                                        try_create_user_json_file)


def initial_install_step_create_user():
    static_user_id = get_static_user_id()

    # If the user has no static install ID, create one
    if static_user_id is None:
        try_create_user_json_file(is_pro=False)
    
    identify()
    log('install_started', {
        'mitoinstaller_version': __version__
    })


def initial_install_step_create_pro_user():
    # Try and create the user.json file, or at least
    # update mitosheet_telemetry if it is not already true
    try_create_user_json_file(is_pro=True)


INITIAL_INSTALLER_STEPS = [
    InstallerStep(
        'Create mito user',
        initial_install_step_create_user
    ),
    InstallerStep(
        'Upgrading mitoinstaller',
        upgrade_mito_installer
    ),
]

INITIAL_INSTALLER_STEPS_PRO = [
    InstallerStep(
        'Create pro mito user',
        initial_install_step_create_pro_user
    ),
    InstallerStep(
        'Upgrading mitoinstaller',
        upgrade_mito_installer
    ),
]






from mitoinstaller.commands import (get_jupyterlab_metadata,
                                    install_pip_packages,
                                    uninstall_pip_packages)
from mitoinstaller.log_utils import log, log_error


def do_upgrade_to_jupyterlab_3():
    """
    Checks that a user has mitosheet installed on JLab 2.0, and
    if so attempts to upgrade them to JLab 3.0 and mitosheet3.

    This is useful for migrating users from JLab 3.0 to JLab 2.0. 

    It does so by:
    1.  Checking we are in the correct state (e.g. 2.0, with mitosheet
        installed.)
    2.  Uninstalling the dependencies of mitosheet and the widget manager
        to not mess with the next step
    3.  Installing mitosheet as normal.

    NOTE: We expect the user to have a user.json if they are going to 
    run this command, as we expect they have used the tool.
    """
    print("Upgrading to JupyterLab 3.0... this may take a moment")
    # First, check that JLab 2.0 is installed, and mitosheet is an extension
    jupyterlab_version, labextensions = get_jupyterlab_metadata()
    if jupyterlab_version is None or not jupyterlab_version.startswith('2'):
        log_error('upgrade_to_jupyterlab_3_failed', {
            'reason': 'Installed jupyterlab version is: {jupyterlab_version}'.format(jupyterlab_version=jupyterlab_version)
        })
        exit(1)
    if (labextensions is not None and len(labextensions) > 2) or ('mitosheet' not in labextensions):
        log_error('upgrade_to_jupyterlab_3_failed', {
            'reason': 'Installed labextensions are: {labextensions}'.format(labextensions=labextensions)
        })
        exit(1)
    
    # First, uninstall mitosheet, jlab
    try:
        uninstall_pip_packages('mitosheet', 'jupyterlab')
    except:
        log_error('upgrade_to_jupyterlab_3_failed', {
            'reason': 'Failed to uninstall mitosheet, jupyterlab'
        })
    # Then install JLab 3.0
    try:
        install_pip_packages('mitosheet3')
    except:
        log_error('upgrade_to_jupyterlab_3_failed', {
            'reason': 'Failed to install mitosheet3'
        })    

    # Log that we finished the upgrade
    log('upgrade_to_jupyterlab_3')



from typing import Callable


class InstallerStep:
    """
    An installer step is a specific step in an install/upgrade process,
    that currently has a name and a function that executes when it runs.

    If you label a step as optional, then although this step may fail,
    it will not stop the InstallerSteps after it from executing.
    """
    def __init__(self, 
        installer_step_name: str, 
        execution_function: Callable,
        optional: bool=False
    ):
        self.installer_step_name = installer_step_name
        self.execution_function = execution_function
        self.optional = optional
    
    def execute(self):
        self.execution_function()



from mitoinstaller.commands import (get_jupyterlab_metadata,
                                    install_pip_packages)
from mitoinstaller.installer_steps.installer_step import InstallerStep


def install_step_mitosheet_check_dependencies():
    jupyterlab_version, extension_names = get_jupyterlab_metadata()

    # If no JupyterLab is installed, we can continue with install, as
    # there are no conflict dependencies
    if jupyterlab_version is None:
        return
    
    # If JupyterLab 2 is installed, then we are are also good to go
    if jupyterlab_version.startswith('2'):
        return
    
    raise Exception('Out of date JupyterLab version {jupyterlab_version}'.format(jupyterlab_version=jupyterlab_version))


def install_step_mitosheet_install_mitosheet():
    install_pip_packages('mitosheet')


def install_step_mitosheet_install_jupyter_widget_manager():
    from jupyterlab import commands
    commands.install_extension('@jupyter-widgets/jupyterlab-manager@2')


def install_step_mitosheet_rebuild_jupyterlab():
    from jupyterlab import commands
    commands.build()


MITOSHEET_INSTALLER_STEPS = [
    InstallerStep(
        'Checking dependencies',
        install_step_mitosheet_check_dependencies
    ),
    InstallerStep(
        'Installing mitosheet',
        install_step_mitosheet_install_mitosheet
    ),
    InstallerStep(
        'Installing @jupyter-widgets/jupyterlab-manager@2',
        install_step_mitosheet_install_jupyter_widget_manager
    ),
    InstallerStep(
        'Rebuilding JupyterLab',
        install_step_mitosheet_rebuild_jupyterlab
    ),
]
